import logging
import os
import shutil
import sys
import uuid
from typing import Any, Dict

import investor8_sdk
import yaml
from mergedeep import merge
from rich.style import Style

from i8_terminal.i8_exception import I8Exception

PACKAGE_PATH = os.path.dirname(__file__)
EXECUTABLE_APP_DIR = os.path.join(os.path.dirname(sys.executable))
OS_HOME_PATH = os.path.expanduser("~")
SETTINGS_FOLDER = os.path.join(OS_HOME_PATH, ".i8_terminal")
METRICS_METADATA_PATH = os.path.join(SETTINGS_FOLDER, "metrics_metadata.csv")
USER_SETTINGS_PATH = os.path.join(SETTINGS_FOLDER, "user.yml")
APP_SETTINGS_PATH = os.path.join(SETTINGS_FOLDER, "config.yml")
ASSETS_PATH = os.path.join(PACKAGE_PATH, "assets")
I8_TERMINAL_LOGO_URL = "https://www.investoreight.com/media/i8t-chart-logo.png"


def init_settings() -> None:
    if not os.path.exists(SETTINGS_FOLDER):
        try:
            os.mkdir(SETTINGS_FOLDER)
        except Exception:
            logging.error(
                f"Cannot initialize app. Application needs write access to create app directory in the following path: \
                    '{OS_HOME_PATH}'"
            )

    if not os.path.exists(USER_SETTINGS_PATH):
        try:
            user_setting = {"app_instance_id": uuid.uuid4().hex}
            with open(USER_SETTINGS_PATH, "w") as f:
                yaml.dump(user_setting, f)
        except Exception:
            logging.error(
                f"Cannot initalize user settings. Make sure you have write access to the path: '{USER_SETTINGS_PATH}'"
            )

    if not os.path.exists(APP_SETTINGS_PATH):
        try:
            app_settings_src_path = os.path.join(PACKAGE_PATH, "config.yml")
            if os.path.exists(app_settings_src_path):
                shutil.copyfile(app_settings_src_path, APP_SETTINGS_PATH)
            else:
                shutil.copyfile(f"{EXECUTABLE_APP_DIR}/config.yml", APP_SETTINGS_PATH)
        except Exception as e:
            logging.error(
                f"Cannot initalize app settings. Make sure you have write access to the path: \
                    '{APP_SETTINGS_PATH}'\n {e}"
            )


def load_user_settings() -> Any:
    if not os.path.exists(USER_SETTINGS_PATH):
        return {}
    with open(USER_SETTINGS_PATH, "r") as f:
        return yaml.safe_load(f) or {}


def load_app_settings() -> Any:
    if not os.path.exists(APP_SETTINGS_PATH):
        return {}
    with open(APP_SETTINGS_PATH, "r") as f:
        return yaml.safe_load(f) or {}


def load_latest_app_settings() -> Any:
    app_settings_src_path = os.path.join(PACKAGE_PATH, "config.yml")
    with open(app_settings_src_path, "r") as f:
        return yaml.safe_load(f) or {}


def save_user_settings(data: Dict[str, Any]) -> None:
    current_user_settings = load_user_settings()
    new_user_settings = {**current_user_settings, **data}
    with open(USER_SETTINGS_PATH, "w") as f:
        yaml.dump(new_user_settings, f)


def delete_user_settings() -> None:
    file = open(USER_SETTINGS_PATH, "w")
    file.close()


def restore_user_settings() -> None:
    current_user_settings = load_user_settings()
    restored_user_setting = {"app_instance_id": current_user_settings.get("app_instance_id")}
    with open(USER_SETTINGS_PATH, "w") as f:
        yaml.dump(restored_user_setting, f)


def find_dicts_diff(dict1: Dict[str, Any], dict2: Dict[str, Any]) -> Dict[str, Any]:
    result = {}
    for k in dict1:
        if k in dict2:
            if type(dict1[k]) is dict:
                res = find_dicts_diff(dict1[k], dict2[k])
                if res:
                    result[k] = res
            if dict1[k] != dict2[k]:
                result[k] = dict1[k]
        else:
            result[k] = dict1[k]
    for k in dict2:
        if k not in dict1:
            result[k] = dict2[k]
    return result


def update_settings() -> None:
    current_app_settings = load_app_settings()
    latest_app_settings = load_latest_app_settings()
    app_new_settings = find_dicts_diff(latest_app_settings, current_app_settings)
    user_new_settings = find_dicts_diff(current_app_settings, latest_app_settings)
    if app_new_settings or user_new_settings:
        new_settings = merge({}, latest_app_settings, current_app_settings)
        with open(APP_SETTINGS_PATH, "w") as f:
            yaml.dump(new_settings, f)


def get_table_style(profile_name: str = "default") -> Dict[str, Any]:
    styles = APP_SETTINGS["styles"]["table"][profile_name]
    try:
        return {
            "header_style": Style(**styles["header"]),
            "row_styles": [Style(**styles["row"]), Style(**styles["alternate_row"])],
            "show_lines": styles.get("show_lines", False),
        }
    except Exception:
        raise I8Exception(
            "Cannot parse table style settings from the configuration file! \
                Check to see if the configuration file is formatted correctly!"
        )


def is_user_logged_in() -> bool:
    if not USER_SETTINGS.get("i8_core_api_key") or not USER_SETTINGS.get("i8_core_token"):
        return False
    return True


def init_api_configs() -> None:
    investor8_sdk.ApiClient().configuration.api_key["apiKey"] = USER_SETTINGS.get("i8_core_api_key")
    investor8_sdk.ApiClient().configuration.api_key["Authorization"] = USER_SETTINGS.get("i8_core_token")
    investor8_sdk.ApiClient().configuration.api_key_prefix["Authorization"] = "Bearer"


def init_notebook() -> None:
    if is_user_logged_in():
        init_api_configs()
    else:
        print("You are not logged in. Please login to i8 Terminal using 'user login' command.")


if "USER_SETTINGS" not in globals():
    init_settings()
    update_settings()
    USER_SETTINGS = load_user_settings()
    APP_SETTINGS = load_app_settings()

if not USER_SETTINGS.get("app_instance_id"):
    user_setting = {"app_instance_id": uuid.uuid4().hex}
    save_user_settings(user_setting)



from typing import Any, Dict, Optional

import click
from rich.console import Console

from i8_terminal.commands.metrics import metrics
from i8_terminal.common.cli import pass_command
from i8_terminal.common.layout import df2Table
from i8_terminal.common.metrics import (
    get_current_metrics_df,
    prepare_current_metrics_formatted_df,
)
from i8_terminal.common.stock_info import validate_tickers
from i8_terminal.common.utils import export_data, export_to_html
from i8_terminal.config import APP_SETTINGS
from i8_terminal.types.metric_identifier_param_type import MetricIdentifierParamType
from i8_terminal.types.ticker_param_type import TickerParamType


@metrics.command()
@click.option(
    "--tickers",
    "-k",
    type=TickerParamType(),
    required=True,
    callback=validate_tickers,
    help="Comma-separated list of tickers.",
)
@click.option(
    "--metrics",
    "-m",
    type=MetricIdentifierParamType(),
    default="pe_ratio_ttm",
    help="Comma-separated list of daily metrics.",
)
@click.option("--export", "export_path", "-e", help="Filename to export the output to.")
@pass_command
def current(tickers: str, metrics: str, export_path: Optional[str]) -> None:
    """
    Lists the given metrics for a given list of companies. TICKERS is a comma-separated list of tickers.
    METRICS can be in the below format:
    {metric}.{period}

    Available periods:
        price, performance and technical metrics:
            1da = 1 day ago,
            2da = 2 days ago

        financial metrics:
            mrq = most recent quarter,
            mry = most recent financial year,
            ttm = trailing 12 months,
            ytd = year to date,
            1qa = 1 quarter ago,
            2qa = 2 quarters ago,
            1ya = 1 year ago,
            2ya = 2 years ago

        earnings metrics:
            uq = upcoming quarter
            mrq = most recent quarter,
            1qa = 1 quarter ago,
            2qa = 2 quarters ago

        You can also use specific period (e.g FY_2020).

    Examples:

    `i8 metrics current --metrics total_revenue.mrq,net_income.2ya,close.1da,total_revenue.FY_2019 \
        --tickers AMD,INTC,QCOM`
    """
    console = Console()
    with console.status("Fetching data...", spinner="material"):
        df = get_current_metrics_df(tickers, metrics)
    if df is None:
        console.print("No data found for metrics with selected tickers", style="yellow")
        return
    for m in [*set(metric.split(".")[0] for metric in set(metrics.split(","))) - set(df["metric_name"])]:
        console.print(f"\nNo data found for metric {m} with selected tickers", style="yellow")
    columns_justify: Dict[str, Any] = {}
    if export_path:
        if export_path.split(".")[-1] == "html":
            for metric_display_name, metric_df in df.groupby("display_name"):
                columns_justify[metric_display_name] = (
                    "left" if metric_df["display_format"].values[0] == "str" else "right"
                )
            table = df2Table(
                prepare_current_metrics_formatted_df(df, "console", include_period=True),
                columns_justify=columns_justify,
            )
            export_to_html(table, export_path)
            return
        export_data(
            prepare_current_metrics_formatted_df(df, "store", include_period=True),
            export_path,
            column_width=18,
            column_format=APP_SETTINGS["styles"]["xlsx"]["financials"]["column"],
        )
    else:
        for metric_display_name, metric_df in df.groupby("display_name"):
            columns_justify[metric_display_name] = "left" if metric_df["display_format"].values[0] == "str" else "right"
        table = df2Table(
            prepare_current_metrics_formatted_df(df, "console", include_period=True), columns_justify=columns_justify
        )
        console.print(table)
